//#define DEBUG 1

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

extern void printk();
extern int sprintf();

extern void pxa_sd_get_response_value(void *);
extern int pxa_sd_opendrain_command(u32, u32);
extern int pxa_sd_put_command(u32, u32);

struct sd_dev {
	struct pxa_sd_busctrl *busctrl;		// 0
	void /*struct sd_control*/ *control;		// 4
	unsigned int __8;					// 8
	unsigned int rca;				// 12

	struct card_info {
		unsigned int block_size;		// 16
		unsigned int write_block_size;		// 20
		unsigned int block_count;		// 24
		unsigned int total_size;		// 28
		unsigned int clock;			// 32

		struct cid {				// 36～88
			char dummy[52];
		} cid;

		struct csd {			// 88 ～ 128
			u8	CSD_STRUCTURE;		// 0
			u8	__1;			// 1
			u8	TAAC;			// 2
			u8	NSAC;			// 3
			u8	TRAN_SPEED;		// 4
			u8	__5;			// 5
			u16	CCC;			// 6,7
			u8	BL_LEN;			// 8
			u8	BL_PARTIAL;		// 9
			u8	WRITE_BLK_MISALIGN;	// 10
			u8	READ_BLK_MISALIGN;	// 11
			u8	DSR_IMP;		// 12
			u8	__13;		// 13
			u16	C_SIZE;			// 14,15
			u8	VDD_R_CURR_MIN;		// 16
			u8	VDD_R_CURR_MAX;		// 17
			u8	VDD_W_CURR_MIN;		// 18
			u8	VDD_W_CURR_MAX;		// 19
			u8	C_SIZE_MULT;		// 20
			u8	SECTOR_SIZE;		// 21
			u8	__22;		// 22
			u8	ERASE_BL_EN;		// 23
			u8	WP_GRP_SIZE;		// 24
			u8	WP_GRP_ENABLE;		// 25
			u8	__26;		// 26
			u8	R2W_FACTOR;		// 27
			u8	WRITE_BL_LEN;		// 28
			u8	WRITE_BL_PARTIAL;	// 29
			u8	__30;			// 30
			u8	FILE_FORMAT_GRP;	// 31
			u8	COPY;			// 32
			u8	PERM_WRITE_PROTECT;	// 33
			u8	TMP_WRITE_PROTECT;	// 34
			u8	FILE_FORMAT;		// 35
			u8	__36;		// 36
			u8	__37;			// 37
			u8	__38;		// 38
			u8	__39;		// 39
		} csd;

		unsigned int	type;				// 128
			// 1 ... mmc
			// 2 ... sd
			// 4 ... sdio
		unsigned int cluster_size;		// 132

		char dummy[20];

		unsigned int	__152;				// 152
	} card_info;				// 16
};

unsigned int ccs = 0;
unsigned int sdhc = 0;

extern struct sd_dev mmcsd_drvr[];

#if DEBUG
static void pxa_sd_get_response_value_mod(unsigned char *val)
{
	char buf[256], *p = buf;
	int i;

	pxa_sd_get_response_value(val);

	for (i = 0; i < 20; i++)
		p += sprintf(p, " %02x", val[i]);

	printk("resp%s\n", buf);
}
#endif

static int pxa_sd_put_command_mod(u32 cmd, u32 arg)
{
#if DEBUG
	printk("%08x: command: %d, 0x%x\n", *(u32 volatile *)0xf8A00010,cmd, arg);
#endif

	if (!sdhc) {
		switch (cmd) {
		case 17:
		case 18:
		case 24:
		case 25:
			arg *= 512;
		}
	}

	return pxa_sd_put_command(cmd, arg);
}



struct pxa_sd_busctrl {
	void (*pxa_sd_reg_open)();
	void (*pxa_sd_reg_release)();
	void (*pxa_sd_reg_reset)();
	void (*pxa_sd_set_clock)(u32, u32);
	void (*pxa_sd_set_transfer_length)();
	void (*pxa_sd_set_transfer_length_multi)();
	void (*pxa_sd_put_command)(u32 cmd, u32 arg);
	void (*pxa_sd_opendrain_command)(u32 cmd, u32 arg);
	int (*pxa_sd_wait_response)();
	int (*pxa_sd_wait_id_response)();
	void (*pxa_sd_get_response_value)(void *);
};
/*
 00(000):   00006850	pxa_sd_reg_open
 04(004):   00006914	pxa_sd_reg_release
 08(008):   000069b0	pxa_sd_reg_reset
 0c(012):   00006b30	pxa_sd_set_clock(?, ?)
 10(016):   00006c60	pxa_sd_set_transfer_length
 14(020):   00006c8c	pxa_sd_set_transfer_length_multi
 18(024):   00006cb8	pxa_sd_put_command
 1c(028):   00007174	pxa_sd_opendrain_command(cmd, arg)
 20(032):   00007188	int pxa_sd_wait_response()
 24(036):   00007368	int pxa_sd_wait_id_response
 28(040):   000074c0	pxa_sd_get_response_value(char *)
 2c(044):   00007520	pxa_sd_wait_complete
 30(048):   0000768c	pxa_sd_receive_chars
 34(052):   0000781c	pxa_sd_transmit_chars
 38(056):   0000797c	pxa_sd_selectcard
 3c(060):   0000799c	pxa_sd_deselectcard
 40(064):   00000000	
 44(068):   00000000	
 48(072):   00000000	
 4c(076):   00007944	pxa_sd_after_transmit
 50(080):   000079bc	pxa_sd_write_protected
 54(084):   00007abc	pxa_sd_card_changed
 58(088):   00006a98	pxa_sd_startclock
 5c(092):   000069c0	pxa_sd_stopclock
 60(096):   00007adc	pxa_sd_set_buswidth
 64(100):   00007af8	pxa_sd_read_direct
 68(104):   00007b1c	pxa_sd_write_direct
 6c(108):   00007b4c	pxa_sd_read_extended
 70(112):   00007c9c	pxa_sd_write_extended
 74(116):   000067a8	pxa_sd_set_power
*/

extern void udelay(u32 us);

static int sd_idle_card_present_sd_mod(struct pxa_sd_busctrl *busctrl, struct card_info *card_info)
{
	unsigned int resp[5];
	int ret;
	int i;

//	printk("sd_idle_card_present_sd_mod\n");

	udelay(100);

	busctrl->pxa_sd_opendrain_command(0, 0);

	udelay(100);


	/* ---> add for SDHC */

	busctrl->pxa_sd_opendrain_command(8, 0x1aa);
	busctrl->pxa_sd_get_response_value(resp);

	/* <--- add for SDHC */


	busctrl->pxa_sd_opendrain_command(55, 0);

	ret = busctrl->pxa_sd_wait_response();
	if (ret == 5)
		return -1;

	busctrl->pxa_sd_opendrain_command(41, 0);

	ret = busctrl->pxa_sd_wait_response();
	if (ret == 5)
		return -1;

	busctrl->pxa_sd_get_response_value(resp);

	for (i = 0; i < 100; i++) {

		udelay(10000);

		busctrl->pxa_sd_opendrain_command(55, 0);

		ret = busctrl->pxa_sd_wait_response();
		if (ret == 5)
			return -1;

		busctrl->pxa_sd_opendrain_command(41, ((resp[0] >> 8) & 0xffffff00) | 0x40000000); // SDHC

		ret = busctrl->pxa_sd_wait_id_response();

		if (ret == 5)
			return -1;

		if (ret > 1)
			continue;

		busctrl->pxa_sd_get_response_value(resp);

		if (resp[1] & 0x80) {
			card_info->type |= 2;

			if (resp[1] & 0x40) {
				ccs = 1;
			} else {
				ccs = 0;
			}

			return 0;
		}
	}

	return -1;
}

static int sd_get_csd_mod(struct pxa_sd_busctrl *busctrl, u32 rca, struct csd *csd)
{
	unsigned int response[5];
	int ret;

	busctrl->pxa_sd_put_command(9, rca << 16);

	ret = busctrl->pxa_sd_wait_response();
	if (ret)
		return ret;

	busctrl->pxa_sd_get_response_value(response);

	csd->CSD_STRUCTURE = response[3] >> 30;
	csd->TAAC = (response[3] >> 16);	// u8
	csd->NSAC = (response[3] >> 8);		// u8
	csd->CCC = response[2] >> 20;	// u16
	csd->BL_LEN = (response[2] >> 16) & 0xf;	// u8
	csd->BL_PARTIAL = (response[2] >> 15) & 0x1;	// u8
	csd->TRAN_SPEED = (response[3] & 0xff);		// u8
	csd->WRITE_BLK_MISALIGN = (response[2] >> 14) & 1;	// u8
	csd->READ_BLK_MISALIGN = (response[2] >> 13) & 1;	// u8
	csd->DSR_IMP = (response[2] >> 12) & 1;	// u8

	if (csd->CSD_STRUCTURE == 0) {
		// CSD ver 1.x
		csd->C_SIZE = ((response[2] << 2) & 0xffc) | (response[1] >> 30);
		csd->VDD_R_CURR_MIN = (response[1] >> 27) & 7;
		csd->VDD_R_CURR_MAX = (response[1] >> 24) & 7;
		csd->VDD_W_CURR_MIN = (response[1] >> 21) & 7;
		csd->VDD_W_CURR_MAX = (response[1] >> 18) & 7;
		csd->C_SIZE_MULT = (response[1] >> 15) & 7;

		sdhc = 0;
	} else {
		// CSD ver 2.0
		csd->C_SIZE = /*((response[2] << 16) & 0x3f) |*/ (response[1] >> 16);
		csd->VDD_R_CURR_MIN = 0;
		csd->VDD_R_CURR_MAX = 0;
		csd->VDD_W_CURR_MIN = 0;
		csd->VDD_W_CURR_MAX = 0;
		csd->C_SIZE_MULT = 8;

		sdhc = ccs;
	}
	csd->ERASE_BL_EN = (response[1] >> 14) & 1;
	csd->SECTOR_SIZE = (response[1] >> 7) & 0x7f;
	csd->WP_GRP_SIZE = (response[1] & 0x7f);
	csd->WP_GRP_ENABLE = response[0] >> 31;
	csd->R2W_FACTOR = (response[0] >> 26) & 7;
	csd->WRITE_BL_LEN = (response[0] >> 22) & 0xf;
	csd->WRITE_BL_PARTIAL = (response[0] >> 21) & 0x1;
	csd->__30 = (response[0] >> 16) & 0x1;
	csd->FILE_FORMAT_GRP = (response[0] >> 15) & 0x1;
	csd->COPY = (response[0] >> 14) & 0x1;
	csd->PERM_WRITE_PROTECT = (response[0] >> 13) & 0x1;
	csd->TMP_WRITE_PROTECT = (response[0] >> 12) & 0x1;
	csd->FILE_FORMAT = (response[0] >> 10) & 0x3;
	csd->__37 = response[0] & 0x7f;

#if DEBUG >= 2
#define print_csd_item(x) printk(#x ": 0x%x (%d)\n", csd->x, csd->x)

	print_csd_item(CSD_STRUCTURE);		// 0
	print_csd_item(__1);			// 1
	print_csd_item(TAAC);			// 2
	print_csd_item(NSAC);			// 3
	print_csd_item(TRAN_SPEED);		// 4
	print_csd_item(__5);			// 5
	print_csd_item(CCC);			// 6,7
	print_csd_item(BL_LEN);			// 8
	print_csd_item(BL_PARTIAL);		// 9
	print_csd_item(WRITE_BLK_MISALIGN);	// 10
	print_csd_item(READ_BLK_MISALIGN);	// 11
	print_csd_item(DSR_IMP);		// 12
	print_csd_item(__13);		// 13
	print_csd_item(C_SIZE);			// 14,15
	print_csd_item(VDD_R_CURR_MIN);		// 16
	print_csd_item(VDD_R_CURR_MAX);		// 17
	print_csd_item(VDD_W_CURR_MIN);		// 18
	print_csd_item(VDD_W_CURR_MAX);		// 19
	print_csd_item(C_SIZE_MULT);		// 20
	print_csd_item(SECTOR_SIZE);		// 21
	print_csd_item(__22);		// 22
	print_csd_item(ERASE_BL_EN);		// 23
	print_csd_item(WP_GRP_SIZE);		// 24
	print_csd_item(WP_GRP_ENABLE);		// 25
	print_csd_item(__26);		// 26
	print_csd_item(R2W_FACTOR);		// 27
	print_csd_item(WRITE_BL_LEN);		// 28
	print_csd_item(WRITE_BL_PARTIAL);	// 29
	print_csd_item(__30);			// 30
	print_csd_item(FILE_FORMAT_GRP);	// 31
	print_csd_item(COPY);			// 32
	print_csd_item(PERM_WRITE_PROTECT);	// 33
	print_csd_item(TMP_WRITE_PROTECT);	// 34
	print_csd_item(FILE_FORMAT);		// 35
	print_csd_item(__36);		// 36
	print_csd_item(__37);			// 37
	print_csd_item(__38);		// 38
	print_csd_item(__39);		// 39
#endif

	{
		u32 block_size = 1 << csd->BL_LEN;
		u32 write_block_size = 1 << csd->WRITE_BL_LEN;
		u32 block_count = (csd->C_SIZE + 1) << (csd->C_SIZE_MULT + 2);
		u32 total_size = block_count >> (10 - csd->BL_LEN);
		printk("block_size: %u\n", block_size);
		printk("write_block_size: %u\n", write_block_size);
		printk("block_count: %u\n", block_count);
		printk("total_size: %ukB\n", total_size);

		printk("SDHC=%d\n", sdhc);
	}

	return ret;
}

static int get_cluster_size_mod(u32 total, u32 type, u32 block_size)
{
	struct temp {
		u32 capacity;
		u32 blocks;
	}
	sd_table [] = {
		{ 0x00800000 / 512, 0x00000010 },	// 8M
		{ 0x40000000 / 512, 0x00000020 },	// 1G   block  512B * 0x20 = 16KB
		{ 0x80000000 / 512, 0x00000040 },	// 2G   block 1024B * 0x40 = 64KB
		{ 0x00000000 / 512, 0x00000020 }	// > 2G block 2048B * 0x20 = 64KB
	},
	mmc_table [] = {
		{ 0x01000000 / 512, 0x00000008 },
		{ 0x08000000 / 512, 0x00000004 },
		{ 0x10000000 / 512, 0x00000008 },
		{ 0x1f800000 / 512, 0x00000010 },
		{ 0x3f000000 / 512, 0x00000020 },
		{ 0x7e000000 / 512, 0x00000040 },
		{ 0x00000000 / 512, 0x00000004 }
	}, *table;

	int index;

	printk("total:%u type:%d block:%d\n", total, type, block_size);

	if (type & 2)
		table = sd_table; // SD
	else
		table = mmc_table; // MMC

	for (index = 0; table[index].capacity; index++) {
		if (total <= table[index].capacity)
			break;
	}

	printk("[SD]-clustsize = %x\n", block_size * table[index].blocks);

	return block_size * table[index].blocks;
}

#if DEBUG >= 2
static int sd_write_single_block_withcheck_mod(struct sd_dev *drvr, u32 pos, char *buff, u32 size)
{
	int ret;
	printk("%s: pos:0x%x size:0x%x\n", __func__, pos, size);
	ret = sd_write_single_block_withcheck(drvr, pos, buff, size);
	printk("%s: ret = %d\n", __func__, ret);
	return ret;
}

static int sd_read_single_block_withcheck_mod(struct sd_dev *drvr, u32 pos, char *buff, u32 size)
{
	int ret;
	printk("%s: pos:0x%x size:0x%x\n", __func__, pos, size);
	ret = sd_read_single_block_withcheck(drvr, pos, buff, size);
	printk("%s: ret = %d\n", __func__, ret);
	return ret;
}
#endif




/* softirq.h is sensitive to the offsets of these fields */
typedef struct {
	unsigned int __softirq_pending;
	unsigned int __local_irq_count;
	unsigned int __local_bh_count;
	unsigned int __syscall_count;
	void * __ksoftirqd_task; /* waitqueue is too large */
} irq_cpustat_t;

extern irq_cpustat_t irq_stat;

#define in_interrupt() ({(irq_stat.__local_irq_count + irq_stat.__local_bh_count != 0);})


#define is_sd_insert() (((*(volatile u32*)0xf8e00000) & 0x200) == 0)
#define MMC_BASE 0xf9100000
#define MMC_STAT (*(volatile u32 *)(MMC_BASE + 0x04))
#define MMC_I_REG (*(volatile u32 *)(MMC_BASE + 0x2c))
#define MMC_CMD (*(volatile u32 *)(MMC_BASE + 0x30))
#define MMC_RXFIFO (*(volatile u32 *)(MMC_BASE + 0x40))

extern volatile u32 jiffies;
extern void schedule_timeout(u32);
extern int pxa_sd_receive_chars(char *buff, int size);

static int pxa_sd_receive_chars_mod(char *buff, int size)
{
	u32 blocks = size / 32;
	u32 cur_size;
	u32 *p = (u32 *)buff;
//	u32 start, lap, end;

//	printk("size:%d\n", size);

	if (size & 0x1f)
		return pxa_sd_receive_chars(buff, size);

//	start = *(u32 volatile *)0xf8A00010;

	while (blocks-- > 0) {
		u32 tmp_counter = 0;
		u32 tmp_jiffies = jiffies;

//		start = *(u32 volatile *)0xf8A00010;

		while ((MMC_I_REG & 0x20) == 0) {
			if (!is_sd_insert())
				return 2;

			if (tmp_counter++ > 0xfffff) {
				printk("pxa_sd_receive_chars:  wait full FIFO time out (cmd=%02d block=%d[32k])\n",
					   MMC_CMD & 0x3f);
				return 2;
			}

			if (jiffies - tmp_jiffies > 49) {

				printk("pxa_sd_receive_chars:  wait full FIFO time out by jiffies (cmd=%02d block=%d[32k])\n",
					   MMC_CMD & 0x3f);
				return 2;
			}
		}

//		lap = *(u32 volatile *)0xf8A00010;

#if 0
		if (size < 32)
			cur_size = size;
		else
			cur_size = 32;

		size -= cur_size;
		cur_size >>= 2;
		cur_size--;
		while (cur_size != -1) {
			cur_size--;
			*p++ = MMC_RXFIFO;
		};
#else
//		printk("cur_size:%d\n", cur_size);

#if 0
		*p++ = MMC_RXFIFO;
		*p++ = MMC_RXFIFO;
		*p++ = MMC_RXFIFO;
		*p++ = MMC_RXFIFO;
		*p++ = MMC_RXFIFO;
		*p++ = MMC_RXFIFO;
		*p++ = MMC_RXFIFO;
		*p++ = MMC_RXFIFO;
#else
		{
			__asm__ __volatile__ (
				"ldr r0, [%1];"
				"ldr r1, [%1];"
				"ldr r2, [%1];"
				"ldr r3, [%1];"
				"str r0, [%0];"
				"str r1, [%0, #4];"
				"str r2, [%0, #8];"
				"str r3, [%0, #12];"
				"ldr r0, [%1];"
				"ldr r1, [%1];"
				"ldr r2, [%1];"
				"ldr r3, [%1];"
				"str r0, [%0, #16];"
				"str r1, [%0, #20];"
				"str r2, [%0, #24];"
				"str r3, [%0, #28];"
				:
				:"r"(p), "r"(MMC_BASE + 0x40)
				:"r0", "r1", "r2", "r3"
				);
			p+=8;
		}
#endif
		size -= 32;
#endif
//		end = *(u32 volatile *)0xf8A00010;
//		printk("lap: %d\n", lap - start);
//		printk("trans: %d\n", end - lap);

		if ((MMC_STAT & 9)) {
			printk("SD read CRC error or timeout\n");
			return 1;
		}
	}

//	end = *(u32 volatile *)0xf8A00010;
//	printk("time:%d\n", end - start);

//	if (!in_interrupt()) {
//		schedule_timeout(1);
//	}

//	end = *(u32 volatile *)0xf8A00010;
//	printk("time:%d\n", end - start);

//	printk("MMC_STAT:%08x\n", MMC_STAT);
	return 0;
}

